//========================================================================================
//
// Basic Playback.c - Demonstrates the Playback interface
//
// Part of the Adobe Premiere 5.0 Plug-In Developer's Toolkit.
//
// Copyright  1992-99, Adobe Systems Incorporated, all rights reserved worldwide.
//
//========================================================================================
#include "stdafx.h"
#include "FrwPlayServer.h"

extern "C" {
#include "prTypes.h"
#include "prPlaymod.h"
}


#ifdef PRMAC_ENV
Fixed MulDiv (Fixed value, long mul, long div);
#endif


#define ERRETURN(val) {if (val) {ATLASSERT(FALSE);return;}}



class CMyMovie
{
public:
	CMyMovie()
	{
		next			= 0;
		RefCnt			= 0;
		data_type		= FrwPlayDataType(0);
	}
	
	~CMyMovie()
	{
		FrwMovie = NULL;
	}

	void Release()
	{
		if (RefCnt > 0)
		{
			RefCnt--;
		} else
		{
			ATLASSERT(FALSE);
		}
	}

	void AddRef()
	{
		RefCnt++;
	}

	static CMyMovie* Create(PCHAR fn)
	{
		CHAR	  sz[_MAX_PATH+1];
		PCHAR	  pszn = NULL;
		DWORD	  fl = GetFullPathName(fn, sizeof(sz), sz, &pszn);
		CMyMovie* tp = Head;
		while (tp != NULL)
		{
			if (_stricmp( sz, tp->FileName) == 0)
			{
				tp->AddRef();
				return tp;
			}
			tp = tp->next;
		}
		tp = new CMyMovie();
		if (tp != NULL)
		{
			CComPtr<IFrwPlayMovie>		JPEGMovie;
			strcpy( tp->FileName, sz );
			if (S_OK == JPEGMovie.CoCreateInstance(CLSID_FrwPlayMovie,NULL,CLSCTX_INPROC_SERVER))
			{
				if (S_OK == JPEGMovie->OpenFile(CComBSTR(sz)))
				{
					if (S_OK == JPEGMovie->GetMovieInfo(NULL,&tp->data_type,&tp->width,&tp->height,&tp->total_time,&tp->frame_rate, &tp->audio_duration, &tp->audio_ratio))
					{
						tp->next = Head;
						Head     = tp;
						tp->AddRef();
						tp->FrwMovie = JPEGMovie;
						return tp;
					} else
					{
						ATLASSERT(FALSE);
					}
				} else
				{
					ATLASSERT(FALSE);
				}
			} else
			{
				ATLASSERT(FALSE);
			}
			delete tp;
		}
		return NULL;
	}

	static void CheckAndFree()
	{
		CMyMovie *tp = Head, *pr = NULL;
		while (tp != NULL)
		{
			if (tp->RefCnt == 0)
			{
				if (pr == NULL)
				{
					Head = tp->next;
					delete tp;
					tp   = Head;
				} else
				{
					pr->next = tp->next;
					delete tp;
					tp   = pr->next;
				}
				continue;
			}
			pr = tp;
			tp = tp->next;
		}
	}

	static void FreeAll()
	{
		CMyMovie* tp;
		while (Head != NULL)
		{
			tp		= Head;
			Head	= tp->next;
			delete tp;
		}
	}


private:
	static CMyMovie*		Head;
	CMyMovie*				next;
	LONG					RefCnt;
public:
	double					total_time;
	double					frame_rate;
	double					audio_duration;
	double					audio_ratio;
	long					width,height;
	FrwPlayDataType			data_type;
	CHAR					FileName[_MAX_PATH+1];
	CComPtr<IUnknown>		FrwMovie;
};

typedef CMyMovie *PCMyMovie;


CMyMovie* CMyMovie::Head = NULL;



class CMyClip
{
private:
	
	CMyClip()
	{
		next		= NULL;
		MyMovie		= NULL;
		with_alpha	= FALSE;
		with_jpeg	= FALSE;
		with_yuv	= FALSE;
		forced_black= FALSE;
	}

public:

	CMyClip(CMyClip* p)
	{
		next			= NULL;
		MyMovie			= p->MyMovie;
		if (MyMovie != NULL)
		{
			MyMovie->AddRef();
		}
		clip_in			= p->clip_in;
		clip_out		= p->clip_out;
		clip_start		= p->clip_start;
		clip_end		= p->clip_end;
		clip_frame_rate	= p->clip_frame_rate;
		with_alpha		= p->with_alpha;
		with_jpeg		= p->with_jpeg;
		with_yuv		= p->with_yuv;
		forced_black	= p->forced_black;
	}

	~CMyClip()
	{
		if (MyMovie != NULL)
		{
			MyMovie->Release();
		}
	}

	static CMyClip* Create(PCHAR fn, double clip_start, double clip_in, double clip_out, double clip_frame_rate)
	{
		CMyClip* tp = new CMyClip();
		if (tp != NULL)
		{
			if ((fn == NULL) || ((tp->MyMovie = CMyMovie::Create(fn)) != NULL))
			{
				if ((tp->MyMovie != NULL) && (clip_out > tp->MyMovie->total_time))
				{
					clip_out = tp->MyMovie->total_time;
				}
				if (clip_in > clip_out)
				{
					clip_in = clip_out;
				}
				tp->clip_in			= clip_in;
				tp->clip_start		= clip_start;
				tp->clip_out		= clip_out;
				tp->clip_end		= clip_start + clip_out - clip_in;
				tp->clip_frame_rate = clip_frame_rate;
				if (tp->MyMovie != NULL)
				{
					tp->width			= tp->MyMovie->width;
					tp->height			= tp->MyMovie->height;
					tp->with_alpha		= (tp->MyMovie->data_type & FrwPlayRLEAlpha) != 0;
					tp->with_jpeg		= (tp->MyMovie->data_type & FrwPlayJPEG) != 0;
					tp->with_yuv		= (tp->MyMovie->data_type & FrwPlayYCrCb) != 0;
				}
				return tp;
			}
			delete tp;
		}
		return NULL;
	}

	static CMyClip* Create(PCHAR fn, double clip_frame_rate)
	{
		CMyClip* tp = CMyClip::Create(fn,0,0,0,clip_frame_rate);
		if (tp != NULL)
		{
			tp->clip_out	= tp->MyMovie->total_time;
			tp->clip_end	= tp->clip_start + tp->MyMovie->total_time;
		}
		return NULL;
	}

	
	CMyClip*	next;
	CMyMovie*	MyMovie;
	double		clip_in;
	double		clip_out;
	double		clip_start;
	double		clip_end;
	double		clip_frame_rate;
	long		width,height;
	BOOL		with_alpha;
	BOOL		with_jpeg;
	BOOL		with_yuv;
	BOOL		forced_black;
};


typedef CMyClip *PCMyClip;



class CPlayList
{
public:
	CPlayList(pmStdParms *stdParms, pmNewListParms *listParms);
	CPlayList(pmStdParms *stdParms, pmPlayOpenParms *openParms);
	~CPlayList();
	void	InitVars();
	void	ShowCurFrame(long new_frame);
	void	playStop();
	void	playPlay(pmPlayParms *playparms);
	void	playSetDisplay(pmDisplayPos *disp);
	void	playPlayIdle();
	int		playGetInfo(pmPlayInfoRec *finfo);
	void	playUpdate();
	void	playClose();
	void	playSetPos(pmSetPosRec *posrec);
	void	playGetPos(pmGetPosRec *posrec);
	int		playStep(pmStepRec *steprec);
	void	playSetPlayrate(int rate);
	void	playSetVol(int vol);
	void	playShuttleBegin(pmShuttleStartRec *shutrec);
	void	playShuttleIdle(int shutspeed);
	void	playListAddVid(pmVidSegRec *vidSeg);
	void	RestartPlay(double inPos, double outPos);
	void	CheckCreateBlackFrame();
	void	TryInitResources();
	void	ReleaseResources();
	static int playNewList (pmStdParms *stdParms, CPlayList **myPriv, pmNewListParms *listParms);
	static int playOpen (pmStdParms *stdParms, CPlayList **myPriv, pmPlayOpenParms *openParms);

private:	
	pmStdParms			m_stdParms;		// standard parms; we save this off during the open call
	pmDisplayPos		m_disp;			// last display record passed to us
	pmPlayTimebase		m_timebase;		// our timebase
	pmPlayParms			m_playParms;	// last playParms we were sent for playmod_Play. Used for playSetPlayrate
	long				m_playID;		// our playback ID (duplicate from openParms)

	double				m_totaltime;	// Full Time of this Clip List
	double				m_frameRate;	// Frame rate of this clip list
	double				m_inPos;		// in time for loop, used for restarting audio when looping
	double				m_outPos;		// out point
	double				m_mrate;		// Time multiplyer
	double				m_playStartPos;	// time in clip list we started playing at

	double				m_shutspeed;	// last shuttle speed
	double				m_shuttime;		// timestamp of last frame we put up

	long				m_frame;		// current frame in our timebase

	BOOL				m_playing;		// are we currently playing?
	BOOL				m_hasvideo;		// this clip has video
	BOOL				m_hasaudio;		// this clip has audio
	BOOL				m_serviceAud;	// call audio service routine during idle
	BOOL				m_loop;			// we're looping

										// Non RealTime Play Started
	BOOL				m_LazyPlayStarted;

	long				m_volume;		// play volume

	CMyClip*			m_ClipsList;	// List of Clips

	long				m_width;
	long				m_height;
	BOOL				m_with_alpha;
	BOOL				m_with_jpeg;
	BOOL				m_with_yuv;
	BOOL				m_only_black;
	BOOL				m_have_resources;

	CComPtr<IFrwPlayServer>		m_PlayServer;	// Our Forward Play Server
	CComPtr<IFrwPlayTimeline>	m_TimeLine;		// Our TimeLine (FIFO) We will put our shots here
	CComPtr<IFrwResourceHolder> m_ResourceHolder;
	CComPtr<IFrwPlayJPEGImage>  m_BlackImage;
	CComPtr<IFrwPlayYCrCbImage> m_YUVBlackImage;
};
typedef CPlayList *PCPlayList;


// Forward Declarations
PREMPLUGENTRY xPlayEntry (int selector, pmStdParms *stdParms, long param1, long param2);
int		playGetIndFormat (pmModuleInfoRec *finfo, int idx);


#define boardNumber	0


// Constructor & Destructor

void CPlayList::InitVars()
{
	m_disp.wind			=   0;	// last display record passed to us
	m_frame				=	-1;	// current frame in our timebase
	m_playing			=	0;	// are we currently playing?
	m_hasvideo			=	0;	// this clip has video
	m_hasaudio			=	0;	// this clip has audio
	m_serviceAud		=	0;	// call audio service routine during idle
	m_mrate				=	1;	// +- playback %
	m_loop				=	0;	// we're looping
	m_playStartPos		=	0;	// frame we started playing at
	m_outPos			=	0;	// out point
	m_inPos				=	0;	// in time for loop, used for restarting audio when looping
	m_volume			=	0;	// play volume
	m_shutspeed			=	0;	// last shuttle speed
	m_shuttime			=	0;	// timestamp of last frame we put up
	m_ClipsList			=   NULL;
	m_LazyPlayStarted	=	FALSE;
	m_only_black		=	TRUE;
	m_have_resources	=	FALSE;

	m_PlayServer.CoCreateInstance(CLSID_FrwPlayServer,NULL,CLSCTX_INPROC_SERVER);
	m_ResourceHolder.CoCreateInstance(CLSID_FrwResourceHolder,NULL,CLSCTX_INPROC_SERVER);
	if ((m_PlayServer != NULL) && (m_ResourceHolder != NULL))
	{
		// We must switch fifo mode to "FrwPlayRealtimeFifo" during normal playback and
		// return to "FrwPlayAllImagesFifo" at the end of playback
		HRESULT hres = m_PlayServer->CreateTimeline(boardNumber,8*1024*1024,4*1024*1024, FrwPlayAllImagesFifo, &m_TimeLine);
		if( hres != S_OK )
		{
			ATLASSERT(FALSE);
			m_PlayServer		= NULL;
			m_ResourceHolder	= NULL;
		}
	} else
	{
		m_PlayServer		= NULL;
		m_ResourceHolder	= NULL;
	}
}


CPlayList::CPlayList(pmStdParms *stdParms, pmNewListParms *listParms)
{
	InitVars();
	m_stdParms	=	*stdParms;							// save off useful stuff
	m_disp		=   listParms->disp;
	m_timebase	=	listParms->listTimebase;			// save the timebase
	m_playID	=	listParms->playID;					// save our playID for callbacks
	m_totaltime =   double(m_timebase.fileDuration) / m_timebase.timeBase;
	m_frameRate =	double(m_timebase.timeBase) / m_timebase.samplesize;
	m_inPos		=	0;
	m_outPos	=	m_totaltime;
	m_hasvideo	=	1;
	m_hasaudio	=	1;
}


CPlayList::CPlayList(pmStdParms *stdParms, pmPlayOpenParms *openParms)
{
	InitVars();
	m_stdParms	=	*stdParms;							// save off useful stuff
	m_disp		=   openParms->disp;
	m_timebase	=	openParms->timebase;				// save off the timebase and duration
	m_playID	=	openParms->playID;					// save our playID for callbacks
	m_totaltime =   double(m_timebase.fileDuration) / m_timebase.timeBase;
	m_frameRate =	double(m_timebase.timeBase) / m_timebase.samplesize;
	m_inPos		=	0;
	m_outPos	=	m_totaltime;
	m_ClipsList =   CMyClip::Create(openParms->theFile.name, m_frameRate);
}


CPlayList::~CPlayList()
{
	m_PlayServer		= NULL;
	m_TimeLine			= NULL;
	m_ResourceHolder	= NULL;

	while (m_ClipsList != NULL)
	{
		CMyClip* tp = m_ClipsList;
		m_ClipsList = tp->next;
		delete tp;
	}
}



int CPlayList::playNewList (pmStdParms *stdParms, CPlayList **myPriv, pmNewListParms *listParms)
{
	*myPriv = new CPlayList(stdParms, listParms);
	if ((*myPriv != NULL) && ((*myPriv)->m_PlayServer == NULL))
	{
		delete *myPriv;
		*myPriv = NULL;
		return playmod_ErrDriver;
	}
	return 0;
}



int CPlayList::playOpen (pmStdParms *stdParms, CPlayList **myPriv, pmPlayOpenParms *openParms)
{
	*myPriv = new CPlayList(stdParms, openParms);
	if ((*myPriv != NULL) && ((*myPriv)->m_PlayServer == NULL))
	{
		delete *myPriv;
		*myPriv = NULL;
		return playmod_ErrDriver;
	}
	return 0;
}





void CPlayList::CheckCreateBlackFrame()
{
	CMyClip* tp		=	m_ClipsList;
	m_only_black	=	TRUE;
	while (tp != NULL)
	{
		if (tp->MyMovie != NULL)
		{
			if (m_only_black)
			{
				m_only_black = FALSE;
				if (m_BlackImage != NULL)
				{
					FrwPlayDataType type;	long width, height, compressed_data_size;
					ERRETURN(S_OK != m_BlackImage->GetImageInfo(&type, &width, &height, &compressed_data_size));
					if ((width != tp->MyMovie->width) || (height != tp->MyMovie->height))
					{
						m_BlackImage	= NULL;
						m_YUVBlackImage	= NULL;
					}
				}
				if (m_BlackImage == NULL)
				{
					BYTE data[768*576*2];
					long data_size = tp->MyMovie->width * tp->MyMovie->height * 2;
					ERRETURN(data_size > sizeof(data));
					for (int i = 0; i < data_size; i += 2)
					{
						data[i] = 0x80; data[i+1] = 0x0;
					}
					CComPtr<IFrwPlayYCrCbImageData> imagedata;
					CComPtr<IFrwPlayYCrCbImage>		image;
					ERRETURN(S_OK != imagedata.CoCreateInstance(CLSID_FrwPlayYCrCbImage,NULL,CLSCTX_INPROC_SERVER));
					ERRETURN(imagedata == NULL);
					imagedata.QueryInterface(&image);
					ERRETURN(S_OK != imagedata->SetImageData(FrwPlayYCrCb, tp->MyMovie->width, tp->MyMovie->height, data, data_size));
					m_BlackImage.CoCreateInstance(CLSID_FrwPlayJPEGImage,NULL,CLSCTX_INPROC_SERVER);
					ERRETURN(m_BlackImage == NULL);
					ERRETURN(S_OK != m_BlackImage->CompressFromImage(image, 0));
					m_YUVBlackImage = image;
				}
				m_width			=	tp->width;
				m_height		=	tp->height;
				m_with_alpha	=	tp->with_alpha;
				m_with_jpeg		=	tp->with_jpeg;
				m_with_yuv		=	tp->with_yuv;
			}
			tp->forced_black = ((m_width		!=	tp->width)		||
								(m_height		!=	tp->height)		||
								(m_with_alpha	!=	tp->with_alpha)	||
								(m_with_jpeg	!=	tp->with_jpeg)	||
								(m_with_yuv		!=	tp->with_yuv));
		}
		tp = tp->next;
	}
}




//---------------------------------------------------------
//	playListAddVid
//
//	When we add a segment, we need to adjust the

void CPlayList::playListAddVid(pmVidSegRec *vidSeg)
{
	double clip_rate		=	vidSeg->rate /100.0;
	double clip_frame_rate	=	(clip_rate*double(vidSeg->fileTimebase.timeBase)) / vidSeg->fileTimebase.samplesize;
	double clip_start		=	vidSeg->inListpos		 / m_frameRate;
	double clip_end			=	(vidSeg->outListpos + 1) / m_frameRate; 
	double clip_in;
	double clip_out;


	m_frame = -1;

	if (clip_frame_rate == 0)
	{
		clip_frame_rate =	m_frameRate;
		clip_in			=	0; 
		clip_out		=	clip_end - clip_start;
	} else
	{
		clip_in			=	double(vidSeg->inFilepos / vidSeg->fileTimebase.samplesize) / clip_frame_rate; 
		clip_out		=	clip_in + (clip_end - clip_start);
	}

	PCMyClip tpn,tpp;

	// Kill/cut all that our new clip overlaps!
	tpn = m_ClipsList;
	tpp = NULL;
	while (tpn != NULL)
	{
		if ((tpn->clip_start >= clip_start) && (tpn->clip_end <= clip_end))
		{
			// Kill this clip
			if (tpp == NULL)
			{
				m_ClipsList = tpn->next;
				delete tpn;
				tpn        = m_ClipsList;
			} else
			{
				tpp->next  = tpn->next;
				delete tpn;
				tpn        = tpp->next;
			}
			continue;
		}
		if ((tpn->clip_start < clip_end) && (tpn->clip_end > clip_start))
		{
			if (tpn->clip_start >= clip_start) 
			{													// cut left part
				tpn->clip_in	= tpn->clip_in + (clip_end - tpn->clip_start);
				tpn->clip_start = clip_end;
			} else 
			if (tpn->clip_end <= clip_end)
			{													// cut right put
				tpn->clip_out   = tpn->clip_out - (tpn->clip_end - clip_start);
				tpn->clip_end	= clip_start;
			} else 
			{													// worst case -- cut into wto clips
				PCMyClip ttt = new CMyClip(tpn);
				ERRETURN(ttt == NULL)

				ttt->next = tpn->next;
				tpn->next = ttt;
				tpn->clip_out   = tpn->clip_out - (tpn->clip_end - clip_start);
				tpn->clip_end	= clip_start;
				ttt->clip_in	= ttt->clip_in + (clip_end - ttt->clip_start);
				ttt->clip_start = clip_end;
				tpn				= ttt;
			}
		}
		tpp = tpn;
		tpn = tpp->next;
	}

	PCMyClip tp;

	if (vidSeg->black)
	{
		// Create and insert black clip in List
		tp =   CMyClip::Create(NULL, clip_start, clip_in, clip_out, clip_frame_rate);
	} else
	{
		// Create and insert real clip in List
		tp =   CMyClip::Create(vidSeg->theFile.name, clip_start, clip_in, clip_out, clip_frame_rate);
	}

	ERRETURN(tp == NULL)

	tpn = m_ClipsList;
	tpp = NULL;
	while (tpn != NULL)
	{
		if (clip_start < tpn->clip_start)
		{
			break;
		}
		tpp = tpn;
		tpn = tpn->next;
	}
	if (tpp == NULL)
	{
		tp->next	=	m_ClipsList;
		m_ClipsList	=	tp;
	} else
	{
		tp->next	=	tpn;
		tpp->next	=	tp;
	}

	if (m_totaltime < clip_end)
	{
		m_totaltime = clip_end;
	}

	CheckCreateBlackFrame();
}


//---------------------------------------------------------------


void CPlayList::TryInitResources()
{
	// check window size
	long x,y,dx,dy;
	HRESULT result = m_PlayServer->GetOutputFormat(boardNumber,&dx,&dy,NULL);
	bool  needWindow;
	ULONG wnd_mask = 0;
	if( dx==m_width && dy==m_height ) 
	{ 
		needWindow=false;  x=0; y=0; 
	} else 
	{
		needWindow = true;
		x = (dx - m_width)  / 2;
		y = (dy - m_height) / 2;
	}
	// lock resources
	CComPtr<IFrwBoard> board;
	result = m_PlayServer->GetBoardInterface(boardNumber,&board);
	ERRETURN( result != S_OK );
	long mask = 0, locked = 0;
	if (m_with_alpha)
	{
		mask		|= rg_bank_alpha_m_;
		wnd_mask	|= wnd_alpha_;
	}
	if (m_with_jpeg || m_with_yuv)
	{
		mask		|= rg_bank_b_m_;
		wnd_mask	|= wnd_b_;
	}
	result = board->LockResources(mask,&locked, m_ResourceHolder);
	ERRETURN( FAILED(result) );
	ERRETURN( mask != locked );
	// check 'in window' mask
	long window_mask;
	result = m_PlayServer->get_InputValue(boardNumber, vdc_input_enabled_window_mask,&window_mask);
	ERRETURN( result != S_OK );
	if( needWindow ) {
		long x0,y0,w,h;
		result = m_PlayServer->GetInputScale(boardNumber, &x0,&y0,&w,&h);
		ERRETURN( result != S_OK );
		if( x0 != x || y0 != y || w != m_width || h != m_height ) 
		{
			result = m_PlayServer->SetInputScale(boardNumber, x,y,m_width,m_height);
			ERRETURN( result != S_OK );
		}
		if( (window_mask & wnd_mask) != wnd_mask ) 
		{
			result = m_PlayServer->put_InputValue(boardNumber, vdc_input_enabled_window_mask, window_mask | wnd_mask);
			ERRETURN( result != S_OK );
		}
	} else if( window_mask & wnd_mask ) 
	{
		result = m_PlayServer->put_InputValue(boardNumber, vdc_input_enabled_window_mask, window_mask & ~wnd_mask);
		ERRETURN( result != S_OK );
	}
	// check JPEG window
	if( m_with_jpeg ) 
	{
		result = m_TimeLine->SetJPEGWindow(x,y,m_width,m_height);
		ERRETURN( result != S_OK );
	}
	// change mixer state for show movie
	result = board->LockResources(rg_mixer_m_, NULL, m_ResourceHolder);
	if( result==S_OK ) {
		CComQIPtr<IFrwMixer> mixer(board);
		ERRETURN( mixer==NULL );
		if( m_with_yuv ) 
		{
			mixer->put_FrwMixerValue(vdc_mixer_b, mixer_b_b_);
		}
		if( m_with_jpeg ) 
		{
			mixer->put_FrwMixerValue(vdc_mixer_b, mixer_b_module_);
		}
		if( m_with_alpha ) 
		{
			mixer->put_FrwMixerValue(vdc_mixer_alpha, mixer_alpha_);
			mixer->put_FrwMixerValue(vdc_mixer_blender, mixer_blender_direct_);
		} else 
		{
			mixer->put_FrwMixerValue(vdc_mixer_blender, mixer_blender_1_);
		}
	}
	board->UnlockResources(rg_mixer_m_,m_ResourceHolder);
	// all right
	m_have_resources = TRUE;
}


void CPlayList::ReleaseResources()
{
	CComPtr<IFrwBoard> board;
	m_have_resources = FALSE;
	HRESULT result = m_PlayServer->GetBoardInterface(boardNumber,&board);
	ERRETURN( result != S_OK );
	board->UnlockResources(0xffffffff,m_ResourceHolder);
}



//--------------------------------------------------------------
// Display the current frame

void CPlayList::ShowCurFrame(long new_frame)
{
	if((m_hasvideo && (new_frame != m_frame)) || (!m_have_resources))
	{
		m_frame = new_frame;
		CComPtr<IFrwPlaySuperClip> Clip;
		CMyClip* tp = m_ClipsList;
		double	frame_time		= m_frame / m_frameRate;
		double	frame_duration;
		HRESULT hres;

		while (tp != NULL)
		{
			frame_duration	= 1.0 / tp->clip_frame_rate;
			if (((frame_time >= tp->clip_start) && (frame_time < tp->clip_end)) ||
				(tp->next == NULL)												|| 
				(tp->next->clip_start > frame_time)
			   )
			{

				if (frame_time + frame_duration > tp->clip_end)
				{
					frame_time = tp->clip_end - frame_duration;
				}
				if (frame_time < tp->clip_start)
				{
					frame_time = tp->clip_start;
				}

				if (m_with_yuv)
				{
					hres = m_PlayServer->get_SuperClip(FrwPlayYCrCb,FrwPlayShowAlphaB,&Clip);
					ERRETURN(hres != S_OK)
					if ((tp->MyMovie != NULL) && !tp->forced_black)
					{
						hres = Clip->AddMovieFragmentWithFrameRate(tp->MyMovie->FrwMovie,tp->clip_frame_rate,
													  frame_time - tp->clip_start + tp->clip_in,
													  frame_duration);
					} else
					{
						hres = Clip->AddYCrCbImage(m_YUVBlackImage, frame_duration);
					}
				} else
				{
					hres = m_PlayServer->get_SuperClip(FrwPlayCompressed,FrwPlayShowAlphaB,&Clip);
					ERRETURN(hres != S_OK)
					if ((tp->MyMovie != NULL) && !tp->forced_black)
					{
						hres = Clip->AddMovieFragmentWithFrameRate(tp->MyMovie->FrwMovie,tp->clip_frame_rate,
													  frame_time - tp->clip_start + tp->clip_in,
													  frame_duration);
					} else
					{
						hres = Clip->AddJPEGImage(m_BlackImage, frame_duration);
					}
				}
				ERRETURN(hres != S_OK)
				break;
			}
			tp = tp->next;
		}
		if (Clip != NULL)
		{
			double restTime;
			hres = m_TimeLine->get_RestTime(&restTime);
			ERRETURN( hres != S_OK )
			if ((!m_LazyPlayStarted) || (restTime > 3.0/m_frameRate))
			{
				playStop();
				ERRETURN( hres != S_OK )
			}

			TryInitResources();
			if (!m_have_resources)
			{
				return;
			}

			hres = m_TimeLine->AddSuperClip(Clip);
			ERRETURN(hres != S_OK)

			if (!m_LazyPlayStarted)
			{
				hres = m_TimeLine->put_State(FrwPlaybackPrepared);
				ERRETURN( hres != S_OK )
				hres = m_TimeLine->put_State(FrwPlaybackPaused);
				ERRETURN( hres != S_OK )
				hres = m_TimeLine->put_State(FrwPlaybackPlaying);
				ERRETURN( hres != S_OK )
				m_LazyPlayStarted = TRUE;
			}
		}
	}
}



//--------------------------------------------------------
// Stop playback

void CPlayList::playStop()
{
	m_playing=0;
	m_serviceAud=0;
	m_LazyPlayStarted = FALSE;
	ReleaseResources();
	HRESULT hres = m_TimeLine->Reset(FrwPlayAllImagesFifo);
//	HRESULT hres = m_TimeLine->Reset(FrwPlayRealtimeFifo);
	ERRETURN( hres != S_OK );
}


//--------------------------------------------------------
// Begin playback
//

void CPlayList::RestartPlay(double inPos, double outPos)
{

// Prep audio
	if(m_playParms.volume && m_hasaudio)
		(m_stdParms.funcs->audioFuncs->audioPlayPrep)(m_playID, m_playParms.startTime, m_playParms.rate, m_playParms.volume);

// show the first frame
	if (inPos > m_totaltime) 
	{
		inPos = m_totaltime - 1 / m_frameRate;
	}
	if (inPos < 0) 
	{
		inPos = 0;
	}
	m_playStartPos  = inPos;
	if (outPos > m_totaltime) 
	{
		outPos = m_totaltime - 1 / m_frameRate;
	}
	if (outPos <= inPos)
	{
		outPos = inPos + 1 / m_frameRate;;
	}
	m_outPos		= outPos;
	ShowCurFrame((LONG)(inPos*m_frameRate+0.4));
	if (!m_have_resources)
	{
		return;
	}

	m_LazyPlayStarted = FALSE;
	HRESULT hres = m_TimeLine->Reset(FrwPlayRealtimeFifo);
	ERRETURN( hres != S_OK )

// Fill Clips
	BOOL FindClips = FALSE;
	CMyClip* tp  = m_ClipsList;
	CMyClip* tpp = NULL;
	CComPtr<IFrwPlaySuperClip> Clip;
	if (m_with_yuv)
	{
		hres = m_PlayServer->get_SuperClip(FrwPlayYCrCb,FrwPlayShowAlphaB,&Clip);
	} else
	{
		hres = m_PlayServer->get_SuperClip(FrwPlayCompressed,FrwPlayShowAlphaB,&Clip);
	}
	ERRETURN(hres != S_OK)

	while (tp != NULL)
	{
		if ((outPos > tp->clip_start) && (inPos < tp->clip_end))
		{
			double start = max(inPos,  tp->clip_start);
			double end   = min(outPos, tp->clip_end);

			if (tpp != NULL)
			{
				double hole_duration = tp->clip_start - tpp->clip_end;
				if (hole_duration != 0.0)
				{
					// ATLASSERT(FALSE);
					Clip->AddEmpty(hole_duration);
					ERRETURN(hres != S_OK)
				}
			}

			FindClips = TRUE;

			if (m_with_yuv)
			{
				if ((tp->MyMovie != NULL) && !tp->forced_black)
				{
					hres = Clip->AddMovieFragmentWithFrameRate(tp->MyMovie->FrwMovie,tp->clip_frame_rate,
												  start - tp->clip_start + tp->clip_in,
												  end - start);
				} else
				{
					hres = Clip->AddYCrCbImage(m_YUVBlackImage, end - start);
				}
			} else
			{
				if ((tp->MyMovie != NULL) && !tp->forced_black)
				{
					hres = Clip->AddMovieFragmentWithFrameRate(tp->MyMovie->FrwMovie,tp->clip_frame_rate,
												  start - tp->clip_start + tp->clip_in,
												  end - start);
				} else
				{
					hres = Clip->AddJPEGImage(m_BlackImage, end - start);
				}
			}
			ERRETURN(hres != S_OK)
			tpp = tp;
		}
		tp = tp->next;
	}


	ERRETURN(!FindClips)

	hres = m_TimeLine->AddSuperClip(Clip);
	ERRETURN(hres != S_OK)


// Start Video
	hres = m_TimeLine->put_State(FrwPlaybackPrepared);
	ERRETURN( hres != S_OK )
	hres = m_TimeLine->put_State(FrwPlaybackPaused);
	ERRETURN( hres != S_OK )
	hres = m_TimeLine->put_State(FrwPlaybackPlaying);
	ERRETURN( hres != S_OK )

	m_playing		= 1;

// start/restart audio
	m_serviceAud=0;
	if(m_playParms.volume && m_hasaudio)
	{
		(m_stdParms.funcs->audioFuncs->audioPlay)(m_playID, m_playParms.startTime, m_playParms.rate, m_playParms.volume);
		m_serviceAud=1;
	}

}




void CPlayList::playPlay(pmPlayParms *playparms)
{
	
//	ATLASSERT(FALSE);
	m_playParms		= *playparms;
	m_volume		= playparms->volume;
	m_inPos			= playparms->inTime / m_frameRate;
	m_outPos		= (playparms->outTime) / m_frameRate;
	m_loop			= playparms->loop;
	m_playStartPos  = playparms->startTime / m_frameRate;
	m_frame			= playparms->startTime;
	m_mrate			= playparms->rate / 100.0;

	if (m_loop)
	{
		RestartPlay(m_playStartPos, m_outPos);
	} else
	{
		RestartPlay(m_playStartPos, m_totaltime);
	}

}


//---------------------------------------------------------
//	playSetDisplay
//
//	Set up the display area

void CPlayList::playSetDisplay(pmDisplayPos *disp)
{
	pmDisplayPos	currentDisp;
	
	currentDisp = m_disp;

	if(disp->wind != currentDisp.wind || disp->originTop != currentDisp.originTop ||
		disp->originLeft != currentDisp.originLeft || disp->dispWidth != currentDisp.dispWidth || disp->dispHeight != currentDisp.dispHeight)
	{
		m_disp = *disp;
	}
}


//-----------------------------------------------
// Idle proc for playback
//
// Give up some time to the video driver
// Service audio as needed

void CPlayList::playPlayIdle()
{
	double		curtime;
	HRESULT		hres;

	hres = m_TimeLine->get_CurrentTime(&curtime);
	ERRETURN(hres != S_OK)

	curtime		= curtime  + m_playStartPos;		// calc the time on cliplist

	ULONG c_f = ULONG(curtime*m_frameRate+0.4);
	ULONG l_f = ULONG(m_outPos*m_frameRate+0.4);
// if we're at the out point
	if(c_f >= l_f)
	{
	// and looping
		if(m_loop)
		{
			RestartPlay(m_inPos, m_outPos);
		}
		else {
			playStop();
		}
	}

	m_frame = LONG(curtime*m_frameRate+0.4);


	// if we're playing..
	if(m_playing)
	{
	// Service the audio if any
		if(m_serviceAud)
			(m_stdParms.funcs->audioFuncs->audioPlayService)(m_playID);
	}
}


//--------------------------------------------------------
// Get an indexed file support format

int playGetIndFormat (pmModuleInfoRec *finfo, int idx)
{
	finfo->playflags = pmFlag_canPlayLists | pmFlag_hwSupport;
		                                    // we'll play playLists
	finfo->classID = 0x06060606L;			// This is our custom classID. It matches the entry
	finfo->filetype = 'AVIV';				// We handle AVI files on Win
											// in our CLSS table resource.
	switch (idx)
	{

		case 0:	finfo->subtype = mmioFOURCC('F', 'R', 'W', 'D');
				break;

		case 1:	finfo->subtype = mmioFOURCC('f', 'r', 'w', 'd');
				break;

/*
		case 0:	finfo->subtype = mmioFOURCC('D', 'W', 'R', 'F');
				break;

		case 1:	finfo->subtype = mmioFOURCC('d', 'w', 'r', 'f');
				break;
*/
		default:return playmod_BadFormatIndex;
	}
	return 0;
}


//---------------------------------------------------------
// Get pertinent file info

int CPlayList::playGetInfo(pmPlayInfoRec *finfo)
{
	m_hasaudio = finfo->hasAudio;
	m_hasvideo = finfo->hasVideo;
	finfo->playsExternal=1;
	finfo->playsProxy=0;
	return 0;
}


//---------------------------------------------------------
// Update the displayed frame.
//

void CPlayList::playUpdate()
{
	ShowCurFrame(m_frame);
}


//--------------------------------------------------------
// Close up this file

void CPlayList::playClose()
{
	delete this;
}


//---------------------------------------------------------
// Set the display time, and optionally play the audio
// associated with that frame

void CPlayList::playSetPos(pmSetPosRec *posrec)
{
	ShowCurFrame(posrec->position);
	if(m_hasaudio && posrec->volume)
		(m_stdParms.funcs->audioFuncs->audioPlayFrame)(m_playID,posrec->position,1,posrec->volume);
}


//---------------------------------------------------------
// Get the playback position, and the playback mode (playing, stopped)

void CPlayList::playGetPos(pmGetPosRec *posrec)
{
	posrec->mode = (m_playing != 0) ? playmode_Playing : playmode_Stopped;
	posrec->position = m_frame;
}


//---------------------------------------------------------
// Step from a frame position to the next one.
// Optionally play back a frame's worth of audio

int CPlayList::playStep(pmStepRec *steprec)
{
	long	newpos, oldpos;

	oldpos = steprec->position;

	if (steprec->rate >= 100)
		newpos = steprec->position + MulDiv(steprec->direction, steprec->rate, 100);
	else
		newpos = steprec->position + steprec->direction;

	if(newpos < 0)
		newpos = 0;
	if(newpos > LONG(m_totaltime*m_frameRate+0.4))
		newpos = LONG(m_totaltime*m_frameRate+0.4);
	if(newpos != oldpos)
	{
		steprec->position = newpos;
		ShowCurFrame(newpos);

		if(m_hasaudio && steprec->volume)
			(m_stdParms.funcs->audioFuncs->audioPlayFrame)(m_playID,oldpos,steprec->direction*steprec->rate,steprec->volume);
	}
	return 0;
}


//--------------------------------------------------------
// Set the playback rate during playback.
// 
// Since we're using host callbacks for playing,
// we just stop and restart at the current position
// and new rate.

void CPlayList::playSetPlayrate(int rate)
{
	pmPlayParms playParms;
	pmGetPosRec posrec;

	if(m_playing)
	{
		playGetPos(&posrec);
		playParms = m_playParms;
		playParms.startTime = posrec.position;
		playParms.rate = rate;
		playStop();
		playPlay(&playParms);
	}
}


//---------------------------------------------------------
// Set the audio volume during playback.
// If we're not playing back, don't bother

void CPlayList::playSetVol(int vol)
{
	if(m_playing && m_serviceAud)
	{
		(m_stdParms.funcs->audioFuncs->audioSetVolume)(m_playID, vol);
		m_volume = vol;
	}
}


//---------------------------------------------------------
// Begin shuttling

void CPlayList::playShuttleBegin(pmShuttleStartRec *shutrec)
{
	m_shutspeed = shutrec->shuttlerate;
	ShowCurFrame(shutrec->position);
	m_shuttime = timeGetTime();
}


//---------------------------------------------------------
// Continue shuttling

void CPlayList::playShuttleIdle(int shutspeed)
{
	long			curms, curtime;

// calc the time in ms from the last frame, scaled by the shuttle rate
	curms = (long)(timeGetTime() - m_shuttime);
	curms = curms * m_shutspeed / 100;

// calc the current playback position, pinned to the movie's bounds
	curtime = m_frame + MulDiv(curms, m_timebase.timeBase, 1000);
	if(curtime < 0)
		curtime = 0;
	if(curtime >= m_timebase.fileDuration)	
		curtime = m_timebase.fileDuration-1;

// show that frame, if we're not already there
	if(curtime != m_frame)
	{
		ShowCurFrame(curtime);
		m_shuttime = timeGetTime();
	}

// save off the new shuttle speed
	m_shutspeed = shutspeed;	
}



//---------------------------------------------------------
// Entry point for the play module

PREMPLUGENTRY xPlayEntry (int selector, pmStdParms *stdParms, long param1, long param2)
{
	int			result = playmod_Unsupported;
	PCPlayList	pl = (PCPlayList)param1;
	
//		ATLASSERT(FALSE);


//	ATLASSERT(FALSE);
	switch(selector)
	{

	case playmod_Startup:
		result = 0;
		break;

	case playmod_Shutdown:
		CMyMovie::FreeAll();
		result = 0;
		break;

	case playmod_GetIndFormat:
		result = playGetIndFormat((pmModuleInfoRec *)param1, param2);
		break;


	
	
	case playmod_Open:
		ATLASSERT(FALSE);
		result = CPlayList::playOpen(stdParms, (PCPlayList*)param1, (pmPlayOpenParms *)param2);
		break;

	case playmod_NewList:
		ATLASSERT(FALSE);
		result = CPlayList::playNewList(stdParms, (PCPlayList*)param1, (pmNewListParms *)param2);
		break;

	case playmod_Close:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playClose();
		CMyMovie::CheckAndFree();
		result = 0;
		break;


	
	
	case playmod_ListAddVid:	// no action needed...
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playListAddVid((pmVidSegRec *)param2);
		result = 0;
		break;

	case playmod_ListEndAdd:
		CMyMovie::CheckAndFree();
		result = 0;
		break;

	case playmod_SetPlayrate:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playSetPlayrate(param2);
		result = 0;
		break;

	case playmod_GetInfo:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		result = pl->playGetInfo((pmPlayInfoRec *)param2);
		break;

	case playmod_PlayIdle:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playPlayIdle();
		result = 0;
		break;

	case playmod_Activate:
		result = 0;
		break;

	case playmod_Update:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playUpdate();
		result = 0;
		break;

	case playmod_SetPos:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playSetPos((pmSetPosRec *)param2);
		result = 0;
		break;

	case playmod_GetPos:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playGetPos((pmGetPosRec *)param2);
		result = 0;
		break;

	case playmod_SetDisp:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playSetDisplay((pmDisplayPos *)param2);
		result = 0;
		break;

	case playmod_Play:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playPlay((pmPlayParms *)param2);
		result = 0;
		break;

	case playmod_Stop:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playStop();
		result = 0;
		break;

	case playmod_Step:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		return pl->playStep((pmStepRec *)param2);
		break;

	case playmod_SetVolume:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playSetVol(param2);
		break;

	case playmod_ShuttleBegin:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playShuttleBegin((pmShuttleStartRec *)param2);
		break;

	case playmod_ShuttleIdle:
		if (pl == NULL) {ATLASSERT(FALSE); return 0;};
		pl->playShuttleIdle(param2);
		break;

	case playmod_ShuttleEnd:
		break;

	}
	return result;
}

